#include "LTFLO.h"

#include "corrs.h"

using namespace cv;


namespace ltflo {

LTFLO::LTFLO(const cv::Mat& image, BBox trackedArea) : frame(0), image(image.clone()), trackedArea(trackedArea), poseMean(trackedArea) {
	scoreSlidingAverage = 0.5;
	imageWindow = BBox(image);
	canonicalFrame = BBox(0, 0, 1, 1);
	
	biPrefLenFine = trackedArea.getShorterSide() / 15.0;
	biPrefLenCoarse = 2 * biPrefLenFine;
	ptWeightsSigma = biPrefLenFine / 2.0;
	
	Pose4 pose = poseMean;
	std::fill_n(poseCov, 16, 0.0);
	poseCov[0 + 4*0] = (3*biPrefLenCoarse) * (3*biPrefLenCoarse);
	poseCov[1 + 4*1] = (3*biPrefLenCoarse) * (3*biPrefLenCoarse);
	poseCov[2 + 4*2] = log(1.2) * log(1.2); // 20% change
	poseCov[3 + 4*3] = (M_PI/9) * (M_PI/9); // 20deg change
	
	if (verbose) {
		std::cerr <<"*** poseMean=" << poseMean<<" trackedArea="<<trackedArea << " imageWindow=" << imageWindow << " biPrefLenFine=" << biPrefLenFine << " biPrefLenCoarse=" << biPrefLenCoarse << " ptWeightsSigma=" << ptWeightsSigma << " pose=" << pose<<"\n";
	}
	
	// create initial feature cloud and initialise edge qual. field
	initialPointsRand(image);
	
	revsVar = revsMean = edgePoints.size()/100.0 * 0.5; // 0.5 %
	revsNum = 1;
	failureFireThr = 1e16; //inf
	
	A = Mat::eye(Size(3,3), CV_64FC1);
	scaleChange = 1;
	
	models.reserve(modelsCount);
	models.push_back(TrackerState(0, image, gr, ptQuality, weightsField, trackedArea, edgePoints, lines));
	
	
}



const BBox& LTFLO::track(const cv::Mat& image) {
	++frame;
	if (verbose) {
		std::cerr << "_frame_" << frame << "\n";
	}
	shiftToOld(); // xxxOld = xxx;
	d.resize(edgePointsOld.size());
	
	this->image = image.clone();
	
	gradientFromBGR8u(image, gr);
	evidenceFromBGR8u(image, gr, evidence);
	
	//If we failed to generate any points last time
	if(edgePoints.size()==0)
	{
		trackedArea = trackedAreaOld;
		trackedArea.invalid=true;
		return trackedArea;
	}
	
	//////////////////////////////////// estimate frame-to-frame transformation /////////////////////
	unsigned ransacTrial, numInliers;
	std::vector<unsigned> correspFull;
	for (ransacTrial = 0; ransacTrial < 10; ++ransacTrial) {
		// coarse transformation
		if (verbose >= 3) {
			std::cerr << "AOld:\n" << AOld << "\n";
		}
		correspFull = projectCorrespondences(AOld, edgePointsOld, image.size());
		guidedEdgeSearchIteration(&correspFull[0], grOld, gr, imageOld, image, a2rot(AOld), biPrefLenCoarse, false, edgePointsOld.size());
		extractLines(lines, correspFull, gr, 1);
		
		// estimates A and other stats
		numInliers = ransacWrapper(A, d, score, optScore,
									linesOld, lines, correspFull, evidence, weightsField, trackedAreaOld, trackedAreaOld,
									ransacThrCoarse, ransacConfCoarse, scaleChange, 0, verbose >= 3, image);
		if (verbose) {
			std::cerr << "R1:" << numInliers << "(" << round(numInliers * 100.0 / edgePointsOld.size()) << "%)|";
		}

		// refined transformation
		correspFull = projectCorrespondences(A, edgePointsOld, image.size());
		unguidedEdgeSearch(&correspFull[0], gr, biPrefLenFine, edgePointsOld.size());
		extractLines(lines, correspFull, gr, 1);

		numInliers = ransacWrapper(A, d, score, optScore,
									linesOld, lines, correspFull, evidence, weightsField, trackedAreaOld, trackedAreaOld,
									ransacThrFine, ransacConfFine, getScaleChange(A), 0, verbose >= 3, image);
		if (verbose) {
			std::cerr << "R2:" << numInliers << "(" << round(numInliers * 100.0 / edgePointsOld.size()) << "%," << 0.01*round(score*100) << "," << round(optScore) << ")|";
			std::cerr << "SAr:" << 0.01*round(score/scoreSlidingAverage*100) << "|";
		}

		trackedArea = projectPts(A, trackedAreaOld);
		
		double poseProb = getPoseProbabilityCircRot(Pose4(trackedArea), poseMean, poseCov);
		if (verbose >= 2) {
			std::cerr << "Prob:" << poseProb << "|";
		}
		if (numInliers > 0 && poseProb >= correctionProbThreshold*correctionProbThreshold) {
			break;
		}
		
	}
	if (ransacTrial == 10) {
		if (verbose) {
			std::cerr << "Warning: No good solution found by RANSAC(s)!\n";
		}
		A = AOld;
		trackedArea = trackedAreaOld;
	}
	if (verbose >= 2) {
		std::cerr << "\nRANSAC succeeded after " << ransacTrial+1 << " trials!\n";
	}
	
	// update the point quality field and derived stuff
	updatePtQualityWrapper(ptQuality, correspFull, linesOld, A, ptQualityDec, evidence);
	double meanModel = extractModelWindow();
	GaussianBlur(ptQuality, weightsField, Size(0,0), ptWeightsSigma);
	weightsField /= meanModel;
	
	
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// C O R R E C T I O N S
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
	double fLog = log2(frame);
	if (ransacTrial == 10 // nothing good found by RANSAC(s)
			|| score/scoreSlidingAverage < reinitFireThr // sudden drop in score
			|| abs(fLog - round(fLog)) < 1e-16 // logarithmically often to boost learning at the start
			) {
		// TODO if the windowing is used to save speed when computing gradient, the Gradient and Evidence need to be recomputed here from the whole image
		revsThisCorrection = 0.0;
		double topScore = -1;
		int topModel = 0;
		Mat topA, topAPrev;
		std::vector<double> topD;
		std::vector<unsigned> topCorrespFull;
		BBox topTrackedArea;
		
		for (unsigned m = 0; m < models.size(); ++m) {
			if (verbose) {
				std::cerr << "\nREP #" << m << "(f" << models[m].getFrameNo() << ',' << 0.1*round(10*models[m].getTimesUsed()) << "):";
			}
			BBox winOld = trackedAreaOld;
			// = aFromCan * aToCan;
			Mat aPrev = u2a(models[m].getTrackedArea(), trackedAreaOld);
			
			// DO ONE CORRECTION
			//TODO!!!
			Mat repA = Mat::eye(3,3, CV_64F);
			std::vector<double> repD;
			std::vector<unsigned> repCorrespFull;
			double repScore = doOneCorrection(repA, repD, repCorrespFull, aPrev, m); // or something like that
			// HERE!!!
			//TODO count line reversions and fire failure if exceeded the distribution (99th percentile)
			
			repA = repA * aPrev.inv();
			BBox repTrackedArea = projectPts(repA, trackedAreaOld);
			double repPoseProb = getPoseProbabilityCircRot(Pose4(repTrackedArea), poseMean, poseCov);
			if (repScore > topScore && repPoseProb >= correctionProbThreshold) {
				// TODO: try to use the probability as a proper (multiplicative) prior in the future
				topScore = repScore;
				topA = repA;
				topAPrev = aPrev;
				topD = repD;
				topCorrespFull = repCorrespFull;
				topTrackedArea = repTrackedArea;
				topModel = m;
			}
			
			models[m].fadeUsage();
		}
		
		// check for number of reversions
		failureFireThr = revsMean + 2.58 * sqrt(revsVar); // 2.58*stdev = 99th percentile of normal distribution
		if (verbose >= 2) {
			std::cerr << "\nfailureFireThr =\n\t" << failureFireThr << "\n";
		}
		++revsNum;
		updateMVNPD(&revsMean, &revsVar, &revsThisCorrection, revsNum, 1);
		
		models[topModel].increaseUsage();
		if (verbose) {
			std::cerr << "Corr.m.:" << models[topModel].getFrameNo() << '|';
		}

		double relShift = trackedArea.getRelativeShift(topTrackedArea);
		if (verbose >= 2) {
			std::cerr << "Rel.shift:" << 0.01*round(100*relShift) << "|";
		}
		if (relShift < shiftThrRepair) { // we're ok, not repairing
			if (verbose) {
				std::cerr << "Tracking :)|";
			}
			if (relShift < shiftThrLearn) { // we're better than ok, learning a new model
				if (verbose) {
					std::cerr << "We have new model!|"; // TODO this may be better to just remember and save it later, when we filter&add edge points...
				}
				TrackerState currentState(frame, image, gr, ptQuality, weightsField, trackedArea, edgePoints, lines);
				addModel(currentState);
			}
			failureFireThr = 1e16; //inf
		} else {
			// condition inverted because of NaNs
			if ( !(topScore <= score * min(30.0, optScore) * sqrt(double(numInliers) / edgePointsOld.size())) ) { // we're worse, performing the correction
				if (verbose) {
					std::cerr << "Correction|";
				}
				A = topA;
				d = topD;
				correspFull = topCorrespFull;
				trackedArea = topTrackedArea;
				ptQuality = models[topModel].getPtQuality();
				updatePtQualityWrapper(ptQuality, correspFull, models[topModel].getLines(), A*topAPrev, ptQualityDec, evidence);
				double meanModel = extractModelWindow();
				GaussianBlur(ptQuality, weightsField, Size(0,0), ptWeightsSigma);
				weightsField /= meanModel;
				
				failureFireThr = 1e16; //inf
			} else { // we're better than the correction, rejecting it!
				if (verbose) {
					std::cerr << "Tracking :(|";
					if (topScore == -1) { // no correction could be accepted!
						std::cerr << "Prior prob. too low!|";
					}
				}
			}
			
			
		}
	}
	
	if (revsThisCorrection >= failureFireThr) {
		if (verbose) {
			std::cerr << "Tracking failure!\n";
		}
//		trackedArea = BBox::invalid();
//		return BBox::invalid();
	}
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// E N D   O F   C O R R E C T I O N S
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
	
	if (verbose >= 3 ) {
		std::cerr << "\nRatio of TA inside = " << round(intersectionArea(trackedArea, imageWindow) / trackedArea.getArea() * 100.0) << "%\n";
	}
	
	scoreSlidingAverage = scoreSlidingAverage * scoreSlidingAverageCoeff + score * (1 - scoreSlidingAverageCoeff);

	
	
	scaleChange = getScaleChange(A);

	biPrefLenFine *= scaleChange;
	biPrefLenCoarse *= scaleChange;
	ptWeightsSigma *= scaleChange;
	asymptoticNumberPoints *= scaleChange; // only linear?
	
	if (biPrefLenFine < 1) // object too small/distant
	{
		if (verbose) {
			std::cerr << "Object too small("<<trackedArea<<")! enlarging to ";
		}
		trackedArea = BBox::BBoxLTWH(trackedArea.l()-10,trackedArea.t()-10,  trackedArea.w()+20, trackedArea.h()+20);
		trackedArea.invalid=true;
		
		if (verbose) {
			std::cerr <<trackedArea<<"\n";
		}
	}
	if(trackedArea.getArea() > image.rows*image.cols) { // object too large/close

		if (verbose) {
			std::cerr << "Object too big("<<trackedArea<<")! shrinking to ";
		}
		trackedArea = BBox::BBoxLTWH(trackedArea.l()+trackedArea.w()/4,trackedArea.t()+trackedArea.h()/4,  trackedArea.w()/2, trackedArea.h()/2);
		trackedArea.invalid=true;
		if (verbose) {
			std::cerr <<trackedArea<<"\n";
		}
	}
	if(intersectionArea(trackedArea, imageWindow) / trackedArea.getArea() < .25) // object out of image
	{
		if (verbose) {
			std::cerr << "Object outside image ("<<trackedArea<<"). setting to middle\n";
		}
		trackedArea.invalid=true;
		//trackedArea = BBox::invalid();
		//return BBox::invalid();
		return trackedArea;
	}
	
	
	// keep inliers & add new edgels
	std::vector<Point> inliers;
	for (unsigned p = 0; p < d.size(); ++p) {
		if (d[p] <= survivorsThr) {
			inliers.push_back(Point(correspFull[4*p+2], correspFull[4*p+3]));
		}
	}
	if (inliers.size() == 0) { // we obviously jumped somewhere wrong
		if (verbose) {
			std::cerr << "Out of inliers!\n";
		}
		trackedArea.invalid=true;
		return trackedArea;
	}
	std::sort( inliers.begin(), inliers.end(), &comparePoints );
	inliers.erase( unique(inliers.begin(), inliers.end()), inliers.end() ); // unique inliers only
	if (verbose) {
		std::cerr << inliers.size() << " inls|";
	}
	
	// add new points, and crop those outside the rectangle
	std::vector<Point> newPoints = addEdgels(asymptoticNumberPoints - inliers.size());
	edgePoints = combineNearPoints(inliers, biPrefLenFine, newPoints, biPrefLenFine/2);
	if (verbose) {
		std::cerr << edgePoints.size() << " edgels remaining\n";
	}

	Pose4 pose(trackedArea);
	updatePoseProbabilityDistribution(pose, frame+1);
	
	return trackedArea;
}


//find a random point inside TA by mixing corners
cv::Point LTFLO::randPointInTA() const
{
	//4 random mixing coefficients
	double weights[4]={rand(), rand(), rand(), rand()};
	double sum=weights[0]+weights[1]+weights[2]+weights[3];
	weights[0]/=sum;
	weights[1]/=sum;
	weights[2]/=sum;
	weights[3]/=sum;
	//do mixing
	cv::Point out = weights[0] * trackedArea[0] + weights[1] * trackedArea[1] + weights[2] * trackedArea[2] + weights[3] * trackedArea[3];
	return out;
}


void LTFLO::initialPointsRand(const cv::Mat& image) {
	// gradient
	gradientFromBGR8u(image, gr);
	if(verbose)
	{
		std::cerr<<"Gradient image size="<<gr.x.size()<<"\n";
		std::cerr<<"Tracked area: L="<<trackedArea.l()<<
		" R="<<trackedArea.r()<<" T="<<trackedArea.t()<<" B="<<
		trackedArea.b()<<"\n";
		std::cerr<<"Generating random points from ="<<trackedArea.l()<<
		","<<trackedArea.t()<<" to "<<trackedArea.l()+trackedArea.w()<<","<<
		trackedArea.t()+trackedArea.h()<<"\n";
	}
	
	// generate random points
	std::vector<unsigned> correspFull(4 * numGenPts); // in the format x1-y1-x2-y2
	for (unsigned point = 0; point < numGenPts; ++point) {
		/*int px = rand() % unsigned(trackedArea.w()) + trackedArea.l();
		int py = rand() % unsigned(trackedArea.h()) + trackedArea.t();
		correspFull[point*4] = correspFull[point*4 + 2] = px;
		correspFull[point*4 + 1] = correspFull[point*4 + 3] = py;*/
		cv::Point p=randPointInTA();
		correspFull[point*4] = correspFull[point*4 + 2] = p.x;
		correspFull[point*4 + 1] = correspFull[point*4 + 3] = p.y;
	}
	// iterate to edges
	unguidedEdgeSearch(&correspFull[0], gr, biPrefLenCoarse, numGenPts);
	// save
	edgePoints.clear();
	edgePoints.reserve(numGenPts);
	if(verbose)
		std::cerr<<"Creating edge quality field of size="<<image.size()<<"\n";
	
	ptQuality = Mat::zeros(image.size(), CV_64FC1);
	for (unsigned point = 0; point < numGenPts; ++point) {
		Point p(correspFull[point*4+2], correspFull[point*4+3]);
		if(pointInsideTA(p, 0)) {

			edgePoints.push_back(p);
			ptQuality.at<double>(p) = 3;
		}
	}
	std::sort( edgePoints.begin(), edgePoints.end(), &comparePoints );
	if(verbose)
		std::cerr<<"sorted "<<edgePoints.size()<<"edge points\n";
	edgePoints.erase( unique(edgePoints.begin(), edgePoints.end()), edgePoints.end() ); // unique edge points only
	if(verbose)
		std::cerr<<"kept "<<edgePoints.size()<<"unique edge points\n";
	if(verbose)
		std::cerr<<"Extracting model window\n";
	// blur the ptQuality field and extract model window
	double meanModel = extractModelWindow();
	if(verbose)
		std::cerr<<"Meanmodel="<<meanModel<<"\n";
	GaussianBlur(ptQuality, weightsField, Size(0,0), ptWeightsSigma);
	weightsField /= meanModel;
	if(verbose)
		std::cerr<<"Blurred\n";
	
	// lines
	extractLines(lines, edgePoints, gr, 1);
	if(verbose)
		std::cerr<<"lines extracted\n";
	
	asymptoticNumberPoints = edgePoints.size() * 3/2;
	if (verbose) {
		std::cerr << "Starting with " << edgePoints.size() << " edgels.\n";
	}
}


void LTFLO::shiftToOld(void) {
	imageOld = image.clone();
	trackedAreaOld = trackedArea;
	AOld = A.clone();
	grOld = gr.clone();
	edgePointsOld = edgePoints;
//	linesOld = lines;
	extractLines(linesOld, edgePointsOld, grOld, 1);
}


//double meanModel = extractModelWindow();
double LTFLO::extractModelWindow(void) {
	Mat AtoCanon = u2a(trackedArea, canonicalFrame);
	Mat smoothWeights;
	double sigma = trackedArea.getDiagonal() / modelWindowSize / 3.0;
	if(verbose)
		std::cerr<<"Blurring with a gaussian of sigma="<<sigma<<" (target diagonal="<<trackedArea.getDiagonal()<<" and modelWindowSize="<<modelWindowSize<<")\n";
	GaussianBlur(ptQuality, smoothWeights, Size(0,0), sigma);
	
	double modelSum = 0;
	
	modelWindow = Mat::zeros(modelWindowSize, modelWindowSize, CV_64F);
	for (unsigned row = 0; row < modelWindowSize; ++row) {
		for (unsigned col = 0; col < modelWindowSize; ++col) {
			Point2d canonicalPt(1.0 / (modelWindowSize - 1) * col, 1.0 / (modelWindowSize - 1) * row);
			Point2d imagePt = projectPt(AtoCanon.inv(), canonicalPt);
			double val = interpolateGrayscale(smoothWeights, imagePt);
			modelWindow.at<double>(row, col) = (val == val ? val : 0); // to check for NANs and replace them with zeros
			modelSum += (val == val ? val : 0);
		}
	}
	
	return modelSum / modelWindowSize / modelWindowSize;
}

// adds new points on edges
std::vector<cv::Point> LTFLO::addEdgels(unsigned numberEdgels) {
	std::vector<Point> res;
	
	if (numberEdgels > numGenPts) { // an unsigned way to say <0, if we underflew
		return res;// but that should never happen as there should never be more inliers than asympt. points!
	}
	
	Mat AfromCanon = u2a(canonicalFrame, trackedArea);
	// generate random points
	std::vector<unsigned> correspFull(4 * numberEdgels); // in the format x1-y1-x2-y2
	int M=gr.m.size().height, N=gr.m.size().width;
	for (unsigned point = 0; point < numberEdgels; ++point) {
		Point2d canonicalPt(drand(), drand());
		Point2d imagePt = projectPt(AfromCanon, canonicalPt);
		correspFull[point*4] = correspFull[point*4 + 2] = min(double(N-1), max(0., round(imagePt.x)));
		correspFull[point*4 + 1] = correspFull[point*4 + 3] = min(double(M-1), max(0., round(imagePt.y)));
	}
	// iterate to edges
	if(verbose>2)
	{
		unsigned minx=99999,miny=99999,maxx=0,maxy=0;
		for (unsigned point = 0; point < numberEdgels; ++point)
		{
			minx=min(minx,correspFull[point*4 + 2]);
			maxx=max(maxx,correspFull[point*4 + 2]);
			miny=min(miny,correspFull[point*4 + 3]);
			maxy=max(maxy,correspFull[point*4 + 3]);
		}
		std::cerr<<"added "<<numberEdgels<<" edgels in range "<<minx<<","<<miny<<" to "<<maxx<<","<<maxy<<", using matrix "<<AfromCanon<<".Performing unguided search using image of size "<<gr.x.size()<<std::endl;
	}
	unguidedEdgeSearch(&correspFull[0], gr, biPrefLenCoarse, numberEdgels);
	
	for (unsigned point = 0; point < numberEdgels; ++point) {
		Point p(correspFull[point*4+2], correspFull[point*4+3]);
		if (p.x >= 0 && p.x < image.cols && p.y >= 0 && p.y < image.rows) {
			res.push_back(p);
		}
	}
	std::sort( res.begin(), res.end(), &comparePoints );
	res.erase( unique(res.begin(), res.end()), res.end() ); // unique edge points only
	
	return res;
}


// combines two sets of points together, discarding those outside of given bounds
std::vector<cv::Point> LTFLO::combineNearPoints(const std::vector<cv::Point>& pts1, const double dist1, const std::vector<cv::Point>& pts2, const double dist2) {
	std::vector<Point> res;
	res.reserve(pts1.size() + pts2.size()); // we assume most of them will pass, so a preallocation might not be a bad idea
	
	// first vector
	for (unsigned p = 0; p < pts1.size(); ++p) {
		Point point = pts1[p];
		if (pointInsideTA(point, dist1)) {
			res.push_back(point);
		}
	}
	// second vector
	for (unsigned p = 0; p < pts2.size(); ++p) {
		Point point = pts2[p];
		if (pointInsideTA(point, dist2)) {
			res.push_back(point);
		}
	}
	
	return res;
}

// checks if a point is inside the trackedArea with a given tolerance
bool LTFLO::pointInsideTA(cv::Point point, double tolerance) {
	Point3d ph(point.x, point.y, 1);
	for (unsigned corner = 0; corner < trackedArea.size(); ++corner) {
		Point2d corner1 = trackedArea[corner];
		Point2d corner2 = trackedArea[(corner+1) % trackedArea.size()]; // they go counter-clockwise!
		Point3d c1h(corner1.x, corner1.y, 1);
		Point3d c2h(corner2.x, corner2.y, 1);
		Point3d line = c1h.cross(c2h);
		line /= line.x*line.x + line.y*line.y; // normalise
		double distance = line.dot(ph); // signed distance, positive outside
		if (distance > tolerance) {
			return false;
		}
	}
	// if not too far outside any edge, keep it!
	return true;
}

// updates the distribution of poses (online Gaussian distr.) based on the N-th observation
void LTFLO::updatePoseProbabilityDistribution(Pose4 pose, unsigned N) {
	updateMVNPD(poseMean.data, poseCov, pose.data, N, 4); // matrix storage should be column-major, but it's symmetric, so it doesn't matter
}

// add a model (TrackerState) into the list, replace the least used one if already full... Returns the position where placed.
unsigned LTFLO::addModel(const TrackerState& state) {
	unsigned replace = 1; // we don't test/replace 0 no matter what, that's our best one forever!
	for (unsigned p = 1; p < modelsCount; ++p) {
		if (models.size() <= p) { // found free space
			models.push_back(state);
			return p;
		}
		if (models[p].getTimesUsed() < models[replace].getTimesUsed()) { // found less used one
			replace = p;
		} else if (models[p].getTimesUsed() == models[replace].getTimesUsed()
					&& models[p].getFrameNo() < models[replace].getFrameNo()) { // if the same, replace the older one
			replace = p;
		}
	}
	models[replace] = state;
	return replace;
}


// performs one correction; not sure yet exactly what it will do though...
double LTFLO::doOneCorrection(cv::Mat& A, std::vector<double>& d, std::vector<unsigned>& correspFull, const cv::Mat& aPrev, unsigned modelNo) {
	double repScore, repOptScore;
	TrackerState& model = models[modelNo]; // this should be const, but in the RANSAC wrapper it would cause stuff
	unsigned LEN = model.getBorder().size();
	d.resize(LEN);
	// coarse estimate
	correspFull = projectCorrespondences(aPrev, model.getBorder(), image.size());
	guidedEdgeSearchIteration(&correspFull[0], model.getGradient(), gr, model.getImage(), image, a2rot(aPrev), biPrefLenCoarse, true, LEN);
	std::vector<double> repLines(3*LEN);
	double repRevs = extractLines(repLines, correspFull, gr, 1);
	revsThisCorrection += repRevs / models.size();
	if (verbose) {
		std::cerr << '_' << repRevs << "_:";
	}
	unsigned numInliers = ransacWrapper(A, d, repScore, repOptScore,
								model.getLines(), repLines, correspFull, evidence, model.getWeightsField(), trackedAreaOld, model.getTrackedArea(),
								ransacThrCoarse, ransacConfCoarse, getScaleChange(aPrev), 1, verbose >= 3, image);
	if (verbose) {
		std::cerr << "R1:" << round(numInliers * 100.0 / LEN) << "%|";
	}
	
	// refined estimate
	correspFull = projectCorrespondences(A, model.getBorder(), image.size());
	unguidedEdgeSearch(&correspFull[0], gr, biPrefLenFine, LEN);
	extractLines(repLines, correspFull, gr, 1);
	numInliers = ransacWrapper(A, d, repScore, repOptScore,
								model.getLines(), repLines, correspFull, evidence, model.getWeightsField(), trackedAreaOld, model.getTrackedArea(),
								ransacThrFine, ransacConfFine, getScaleChange(A), 1, verbose >= 3, image);
	if (verbose) {
		std::cerr << "R2:" << numInliers << "(" << round(numInliers * 100.0 / LEN) << "%," << 0.01*round(repScore*100) << "," << round(repOptScore) << ")|";
	}
	
	if (modelNo == 0) {
		return repScore * repOptScore * sqrt(double(numInliers) / LEN) * scoreMultip1;
	} else {
		return repScore * repOptScore * sqrt(double(numInliers) / LEN) * scoreMultipO;
	}
}





} // namespace ltflo





