


#include "cvutils.h"
#include "TrackerState.h"
#include "ransac.h"


namespace ltflo {


class LTFLO {
public:
	// empty, just for vectors etc.
	LTFLO(void) {}
	// initialise on the first frame
	LTFLO(const cv::Mat& image, BBox trackedArea);
	// track a frame!
	const BBox& track(const cv::Mat& image);

	const BBox& getTrackedArea() const {
		return trackedArea;
	}



private:
	// find new points randomly distributed over the target area, and initialise edge quality field and related stuff
	void initialPointsRand(const cv::Mat& image);
	// just a simple copy from xxx to xxxOld, including clone()-s for cv::Mat-s
	inline void shiftToOld(void);
	// extracts the model window and returns its mean value
	double extractModelWindow(void);
	// adds new points on edges
	std::vector<cv::Point> addEdgels(unsigned numberEdgels);
	// combines two sets of points together, discarding those outside of given bounds
	std::vector<cv::Point> combineNearPoints(const std::vector<cv::Point>& pts1, const double dist1, const std::vector<cv::Point>& pts2, const double dist2);
	// checks if a point is inside the trackedArea with a given tolerance
	bool pointInsideTA(cv::Point point, double tolerance = 0);
	// updates the distribution of poses (online Gaussian distr.) based on the N-th observation
	void updatePoseProbabilityDistribution(Pose4 pose, unsigned N);
	// performs one correction; not sure yet exactly what it will do though...
	double doOneCorrection(cv::Mat& A, std::vector<double>& d, std::vector<unsigned>& correspFull, const cv::Mat& aPrev, unsigned modelNo);
	// add a model (TrackerState) into the list, replace the least used one if already full... Returns the position where placed.
	unsigned addModel(const TrackerState& state);
	cv::Point randPointInTA() const;

private:
	// frame id
	unsigned frame;
	// frame data
	cv::Mat image, imageOld;
	// the target location
	BBox trackedArea, trackedAreaOld;
	// frame-to-frame transformation
	cv::Mat A, AOld;
	// scale change in the last frame
	double scaleChange;
	// point errors
	std::vector<double> d;
	// tracking quality
	double score, optScore;

	// gradient, edge points (aka border), lines
	Gradient gr, grOld;
	std::vector<cv::Point> edgePoints, edgePointsOld;
	std::vector<double> lines, linesOld; // a1,b1,c1, a2,b2,c2,...
	Evidence evidence;
	// edge quality field, the same blurred for weighting, and a cut-out window
	cv::Mat ptQuality, weightsField, modelWindow;

	// score sliding average
	double scoreSlidingAverage;
	// pose distribution
	Pose4 poseMean;
	double poseCov[16];
	// distribution of line direction reversions
	double revsThisCorrection;
	double revsMean;
	double revsVar;
	unsigned revsNum;
	// stored "models", aka older tracker states, for corrections
	std::vector<TrackerState> models;
	
	
	

///////////////// PSEUDO-STATIC PSEUDO-CONST SERVICE STUFF /////////////////
private:
	// bbox of the whole image (corner to corner)
	BBox imageWindow;
	// corners of a unit square
	BBox canonicalFrame;
	// edge search (also border iteration) preferred length
	double biPrefLenCoarse, biPrefLenFine;
	// blurring extent for point weighting
	double ptWeightsSigma;
	// the number of points generated in later frames (automatically estimated)
	unsigned asymptoticNumberPoints;
	// threshold to flag tracking failure (global corrections, or report failure up)
	double failureFireThr;


/////////////////////// STATIC SERVICE STUFF //////////////////////////////
public:
	// verbosity level, 0 = nothing, 3 = full
	static const unsigned verbose = 1;
	// how many points to start with?
	static const unsigned numGenPts = 3000;
	// blurring parameters
	static const double sigma = 0.8;
	static const unsigned kernelSize = 7;
	// RANSAC settings
	static const double ransacThrCoarse = 10;
	static const double ransacThrFine = 7;
	static const double ransacConfCoarse = 1 - 1e-2;
	static const double ransacConfFine = 1 - 1e-2;
	// which points to keep? (inliers)
	static const double survivorsThr = 10; //=ransacThrCoarse
	// how many saved states to keep?
	static const unsigned modelsCount = 5;
	// fading factor for score sliding average
	static const double scoreSlidingAverageCoeff = 0.9;
	//fading factor for edge quality field
	static const double ptQualityDec = 0.9;
	// how bad the score drop or current probability needs be to start a correction
	static const double reinitFireThr = .9;
	static const double correctionProbThreshold = 1e-15;
	// how close to each other the best correction and current solution need to be to skip the correction...
	static const double shiftThrRepair = .06;
	// ...and to save the current solution (relative to the bbox shorter side)
	static const double shiftThrLearn = .03;
	// score multipliers - a bit of black magic. The correction is always weaker, and we want to bias in favour of the first one
	static const double scoreMultip1 = 4;
	static const double scoreMultipO = 3;
	static const double scoreMultip1r = 3;
	static const double scoreMultipOr = 2.5;
	// size of the "model" (ie pt. qual. field) window
	static const double modelWindowSize = 64;

};


}





