/* Library for handling with correspondences, mostly 1D edge searches... */

#include "geometry.h"
#include "cvutils.h"


namespace ltflo {

/* Full unguided edge search */
/* function [ newBorder, corresp ] = iterateBorder( border, gm, ga, prefLen) */
/*
Parameters:
correspFull  I/O  4xLEN    on input holds correspondences from past to i_t-1, on output from past to i_t.
                           (Transitive with non-unique entries, if I(A->K,B->K) and K-*X then O(A->X,B->X),
                           if I(A->K,B->L) and K-*X,L-*X then O(A->X,B->X), '->' is in/out corresp. entry and '-*' newly found corresp.)
gr            I   4(MxN)   gradient of the image
prefLen       I   scalar   maximal preferred length
LEN           I   scalar   number of correspondences to search
*/
void unguidedEdgeSearch(unsigned *correspFull,
						const Gradient gr, double prefLen,
						unsigned LEN);

/* One iteration of guided edge search */
/*function [ newBorder, corresp ] = iterateBorderSmart(...
					correspFull, grOld, grNew, imOld, imNew, rot, prefLen, repair)*/
/*
Parameters:
correspFull     I/O     4xLEN    on input holds correspondences from past to it-1, on output from past to it.
                                 (Transitive with non-unique entries, if I(A->K,B->K) and K-*X then O(A->X,B->X),
                                 if I(A->K,B->L) and K-*X,L-*X then O(A->X,B->X), '->' is in/out corresp. entry and '-*' newly found corresp.)
grOld, grNew     I      4(MxN)   gradient of the image
imOld, imNew     I      CCx(M*N) the image, linearised
rot              I      scalar   rotation of the expected motion
prefLen          I      scalar   maximal preferred length
repair           I      0/1 s.   flag, whether we are in a correction or not (affects orientation of lines, not oriented on repair)
M, N, CC, LEN    I      scalars  image dimensions, image color channels, number of correspondences to search
*/
void guidedEdgeSearchIteration(unsigned *correspFull,
								const Gradient grOld, const Gradient grNew, const cv::Mat& imOld, const cv::Mat& imNew,
								double rot, double prefLen, bool repair, unsigned LEN);


/* Measurement of appearance similarity */
/* function [sim] = getAppearanceSimilarity(a1, a2) */
/*
Parameters:
a1, a2  I CC*(2*halfLen+1) measured appearances (pixel arrays)
halfLen I scalar           length of the array to each side
CC      I scalar           number of color channels in the images
Returns:
similarity indicator
*/
double appearanceSimilarity(double *a1, double *a2, unsigned halfLen, unsigned CC);


/* Load a pixel array representing an appearance of 1D window */
/* function [app] = getAppearance(img, px, py, pa, len, ih, iw) */
/*
Parameters:
app       O  CC*(2*len+1)  output array for appearance pixel values
im        I  MxNxCC        image to get appearance from
px, py    I  scalar        coordinates of the centre of the window
pa        I  scalar        window direction (along the gradient, perp. to the edge)
len       I  scalar        half-length of the window
M, N, CC  I  scalar        image dimensions and color channels count
*/
void loadAppearance(double *app, const cv::Mat& im, unsigned px, unsigned py, double pa, double len);

// gets lines in homogeneous coordinates at specified locations, based on gradient
// pointShift is an offset (in px), added to all point coordinates when creating lines.
// This allows to create in 1-based lines from 0-based points (if set to 1)
// returns the number of reversions
unsigned extractLines(std::vector<double>& lines, const std::vector<cv::Point>& edgePoints, const Gradient& gr, double pointShift = 0);
unsigned extractLines(std::vector<double>& lines, const std::vector<unsigned>& correspFull, const Gradient& gr, double pointShift = 0);


} // namespace ltflo


