#include "cvutils.h"

#include <fstream>

#include "algebra.h"

using namespace cv;

namespace ltflo {



// get distance between those, relative to the smaller size
double BBox::getRelativeShift(const BBox& other) const {
	double scale = (getShorterSide() + other.getShorterSide()) / 2.0;
	double rmsDist = 0;
	for (unsigned corner = 0; corner < size(); ++corner) {
		double d = norm(operator[](corner) - other[corner]);
		rmsDist += d*d;
	}
	rmsDist = sqrt(rmsDist / size());
	return rmsDist / scale;
}

// get overlap of two areas
double BBox::overlap(const BBox& other) const {
	double bbIntersection = intersectionArea(*this, other);
	double bbUnion = getArea() + other.getArea() - bbIntersection;
	return bbIntersection / bbUnion;
}

void BBox::make_counter_clockwise()
{
	if(!is_counter_clockwise())
	{
		std::cerr<<"******* SWITCHING FROM CLOCKWISE TO CCW BBOX\n";
		cv::Point2d tmp[4];
		for(int i=0;i<4;i++)
			tmp[i]=(*this)[i];
		for(int i=0;i<4;i++)
			(*this)[i]=tmp[3-i];
	}
//	std::cerr<<"******* FINISHED MAKING CCW\n";
}

//https://stackoverflow.com/questions/1165647/how-to-determine-if-a-list-of-polygon-points-are-in-clockwise-order
bool BBox::is_counter_clockwise() const
{
//		std::cerr<<"******* checking clockwiseness\n";
	double clockwiseness=0;
	//take each centre-point of an arc
	for(int i=0; i<4; i++)
	{
		int before=(i-1+4)%4;
//		std::cerr<<"******* edge "<<before<<" to "<<i<<"\n";
		//compute for this edge
		clockwiseness += ((*this)[i].x - (*this)[before].x) / ((*this)[i].y +  (*this)[before].y);
	}
	return clockwiseness<0;
}

std::ostream& operator<<(std::ostream& stream, const BBox& bb) {
	stream << "[" << bb[0] << ", " << bb[1] << ", " << bb[2] << ", " << bb[3] << "]";
	return stream;
}
	
std::ostream& operator<<(std::ostream& stream, const Pose4& pose) {
	stream << "[" << pose[0] << ", " << pose[1] << ", " << pose[2] << ", " << pose[3] << "]";
	return stream;
}


Vec3b interpolateColour(const cv::Mat& image, const Point2d& point) {
	assert(!image.empty());
	assert(image.channels() == 3);

	int x = (int)point.x;
	int y = (int)point.y;

	int x0 = max(0, min(image.cols-1, x));
	int x1 = min(image.cols-1, x);
	int y0 = max(0, min(image.rows-1, y));
	int y1 = min(image.rows-1, y);

	double cx1 = point.x - x;
	double cx0 = 1 - cx1;
	double cy1 = point.y - y;
	double cy0 = 1 - cy1;

	uchar b = (uchar)round(image.at<cv::Vec3b>(y0, x0)[0] * cy0 * cx0
						 + image.at<cv::Vec3b>(y0, x1)[0] * cy0 * cx1
						 + image.at<cv::Vec3b>(y1, x0)[0] * cy1 * cx0
						 + image.at<cv::Vec3b>(y1, x1)[0] * cy1 * cx1);
	uchar g = (uchar)round(image.at<cv::Vec3b>(y0, x0)[1] * cy0 * cx0
						 + image.at<cv::Vec3b>(y0, x1)[1] * cy0 * cx1
						 + image.at<cv::Vec3b>(y1, x0)[1] * cy1 * cx0
						 + image.at<cv::Vec3b>(y1, x1)[1] * cy1 * cx1);
	uchar r = (uchar)round(image.at<cv::Vec3b>(y0, x0)[2] * cy0 * cx0
						 + image.at<cv::Vec3b>(y0, x1)[2] * cy0 * cx1
						 + image.at<cv::Vec3b>(y1, x0)[2] * cy1 * cx0
						 + image.at<cv::Vec3b>(y1, x1)[2] * cy1 * cx1);

	return cv::Vec3b(b, g, r);
}

double interpolateGrayscale(const cv::Mat& image, const Point2d& point) {
	assert(!image.empty());
	assert(image.channels() == 1);
	assert(image.type() == CV_64F);
	
	int x = (int)point.x;
	int y = (int)point.y;

	int x0 = max(0, min(image.cols-1, x));
	int x1 = min(image.cols-1, x0 + 1);
	int y0 = max(0, min(image.rows-1, y));
	int y1 = min(image.rows-1, y0 + 1);

	double cx1 = point.x - x;
	double cx0 = 1 - cx1;
	double cy1 = point.y - y;
	double cy0 = 1 - cy1;
	
	return image.at<double>(y0, x0) * cy0 * cx0
		 + image.at<double>(y0, x1) * cy0 * cx1
		 + image.at<double>(y1, x0) * cy1 * cx0
		 + image.at<double>(y1, x1) * cy1 * cx1;
}


void gradientFromBGR8u(const cv::Mat& image, Gradient& gr) {
	cv::Mat image2;
	cv::cvtColor(image, image2, CV_BGR2GRAY); // grayscale
	image2.convertTo(image2, CV_64F); // double
	image2 /= 255; // [0-1]
	cv::GaussianBlur(image2, image2, cv::Size(7,7), 0.8); // smooth
	image2 = image2.t();
	//gradient(s)
	double* imgD = (double*) image2.ptr();
	Mat gx(image2.clone()), gy(image2.clone()), ga(image2.clone()), gm(image2.clone());
	GradientD gd;
	gd.x = (double*) gx.ptr();
	gd.y = (double*) gy.ptr();
	gd.a = (double*) ga.ptr();
	gd.m = (double*) gm.ptr();
	unsigned limits[4] = {1, 1, image.cols, image.rows}; // ltrb
	getGradient(gd, imgD, image.rows, image.cols, limits);
	gr.x = gx.t();
	gr.y = gy.t();
	gr.a = ga.t();
	gr.m = gm.t();
//	cv::Sobel(image2, gr.x, CV_64F, 1, 0, 3);
//	cv::Sobel(image2, gr.y, CV_64F, 0, 1, 3);
//	gr.m = gr.x.mul(gr.x) + gr.y.mul(gr.y);
//	cv::sqrt(gr.m, gr.m);
//	gr.a.create(gr.m.size(), gr.m.type());
//	cv::MatConstIterator_<double> itX, itY;
//	cv::MatIterator_<double> itA;
//	for (itX = gr.x.begin<double>(), itY = gr.y.begin<double>(), itA = gr.a.begin<double>();
//			itX != gr.x.end<double>();
//			++itX, ++itY, ++itA) {
//		*itA = atan2(*itY, *itX);
//	}
//	namedWindow("gr a");
//	imshow("gr a", gr.a);
//	waitKey(0);
}


void evidenceFromBGR8u(const cv::Mat& image, const Gradient& gr, Evidence& ev) {
	double CannyLoThr = 15;
	cv::Mat image2, labels;
	cv::cvtColor(image, image2, CV_BGR2GRAY); // grayscale
	GaussianBlur(image2, image2, Size(13,13), 2); // blur (emulating Matlab)
	Canny(image2, image2, CannyLoThr, CannyLoThr*2);
	image2 = 255 - image2;
	distanceTransform(image2, ev.d, labels, DIST_L2, DIST_MASK_5, DIST_LABEL_PIXEL);
	
	std::vector<Point> labelLocation;
	labelLocation.push_back(Point(-1,-1));
	for (unsigned row = 0; row < image.rows; ++row) {
		for (unsigned col = 0; col < image.cols; ++col) {
			if(ev.d.at<int>(row,col) == 0) {
				labelLocation.push_back(Point(col,row));
			}
		}
	}
	
	ev.d.convertTo(ev.d, CV_64F);
	ev.d = 1.0 / (ev.d + 1);
	
	ev.a = Mat::zeros(gr.a.size(), CV_64F);
	for (unsigned row = 0; row < image.rows; ++row) {
		for (unsigned col = 0; col < image.cols; ++col) {
			unsigned label = labels.at<int>(row, col);
			Point nearest = labelLocation[label];
			ev.a.at<double>(row, col) = gr.a.at<double>(nearest.y, nearest.x);
		}
	}
	
//	namedWindow("a");
//	imshow("a", ev.d);
//	waitKey(0);
}


double getPoseProbabilityCircRot(Pose4 pose, const Pose4& poseMean, const double * poseCov) {
	double pdf = normpdfN(pose.data, poseMean.data, poseCov, 4);
	pose[3] += 2 * M_PI; // plus 2 PI
	pdf += normpdfN(pose.data, poseMean.data, poseCov, 4);
	pose[3] -= 4 * M_PI; // minus 2 PI
	pdf += normpdfN(pose.data, poseMean.data, poseCov, 4);
	return pdf;
}

bool comparePoints(const cv::Point& p1, const cv::Point& p2) {
	if (p1.x < p2.x) {
		return true;
	} else if (p1.x > p2.x) {
		return false;
	} else {
		return p1.y < p2.y;
	}
}




} // namespace ltflo



