#ifndef __CVUTILS_H__
#define __CVUTILS_H__

#include "opencv2/opencv.hpp"



namespace ltflo {

// image gradient, 1-channel matrices, x,y- derivatives, magnitude and angle
struct Gradient {
	cv::Mat x, y, m, a;
	Gradient clone() const {
		Gradient gr;
		gr.x = x.clone();
		gr.y = y.clone();
		gr.a = a.clone();
		gr.m = m.clone();
		return gr;
	}
};

// image evidence, inverse distance transform (1/1+d) and nearest angle
struct Evidence {
	cv::Mat d, a;
};


// a bounding box, assumed rectangular, possibly rotated
struct BBox {
	cv::Point2d p0, p1, p2, p3;
	bool invalid;
private:
//	static BBox invalidBox; // a unique super-invalid instance
	
public:
	// ctors and the sort
	BBox(void) {invalid=false;}
	BBox(cv::Point2d& p0, cv::Point2d& p1, cv::Point2d& p2, cv::Point2d& p3) : p0(p0), p1(p1), p2(p2), p3(p3)
	{
		invalid=false;
		make_counter_clockwise();
	}
	BBox(double left, double top, double right, double bottom)
		: p0(left, top), p1(left, bottom), p2(right, bottom), p3(right, top)
	{
		invalid=false;
		make_counter_clockwise();
	}
	BBox(const cv::Mat image) {
		invalid=false;
		unsigned iw = image.size().width;
		unsigned ih = image.size().height;
		p0 = cv::Point2d(0, 0);
		p1 = cv::Point2d(0, ih);
		p2 = cv::Point2d(iw, ih);
		p3 = cv::Point2d(iw, 0);
		make_counter_clockwise();
	}
	static BBox BBoxLTWH(double left, double top, double width, double height) {
		return BBox(left, top, left+width, top+height);
	}
	static BBox BBoxLTRB(double left, double top, double right, double bottom) {
		return BBox(left, top, right, bottom);
	}
	
	bool operator==(const BBox& other) {
		if (this == &other) {
			return true;
		}
		return p0 == other.p0 && p1 == other.p1 && p2 == other.p2 && p3 == other.p3;
	}
	
	// a rectangle is valid only if there is no NAN
	bool isNan() const {
		return ! (p0 == p0 && p1 == p1 && p2 == p2 && p3 == p3);
	}
	bool isInvalid() const {
		return invalid;;
	}
	
	// print into an image
	void print(cv::Mat& image, cv::Scalar colour, int lineWidth = 1) const {
		line(image, p0, p1, colour, lineWidth);
		line(image, p1, p2, colour, lineWidth);
		line(image, p2, p3, colour, lineWidth);
		line(image, p3, p0, colour, lineWidth);
	}
	friend std::ostream& operator<<(std::ostream& stream, const BBox& pose);
	// get the sizes
	// it's rectangle, so we have only the choice of two dimensions!
	double getShorterSide() const {
		return std::min(cv::norm(p0 - p1), cv::norm(p1 - p2));
	}
	double getLongerSize() const {
		return std::max(cv::norm(p0 - p1), cv::norm(p1 - p2));
	}
	double getDiagonal() const {
		return cv::norm(p0 - p2);
	}
	double getArea() const {
		return cv::norm(p0 - p1) * cv::norm(p1 - p2);
	}
	// the following expect axis-aligned BBox, and will NOT be right when rotated!
	cv::Size size2D() const { return cv::Size(norm(p1 - p2), cv::norm(p0 - p1)); }
	double h() const { return cv::norm(p0 - p1); } // height, width, top, left, right, bottom...
	double w() const { return cv::norm(p1 - p2); }
	double t() const { return p0.y; }
	double l() const { return p0.x; }
	double r() const { return p2.x; }
	double b() const { return p2.y; }
	cv::Rect rect() const {
		return cv::Rect(p0.x, p0.y, w(), h());
	}
	
	// for a nicer access
	cv::Point2d& operator[](unsigned index) {
		switch (index) {
			case 0: return p0;
			case 1: return p1;
			case 2: return p2;
			case 3: return p3;
			default: throw "Error: index out of bounds!";
		}
	}
	const cv::Point2d& operator[](unsigned index) const {
		switch (index) {
			case 0: return p0;
			case 1: return p1;
			case 2: return p2;
			case 3: return p3;
			default: throw "Error: index out of bounds!";
		}
	}
	unsigned size(void) const {
		return 4;
	}
	
	// copy the box into an array of doubles
	void toDouble(double * arr) const {
		for (unsigned corner = 0; corner < size(); ++corner) {
			arr[2*corner + 0] = operator[](corner).x;
			arr[2*corner + 1] = operator[](corner).y;
		}
	}
	
	// get distance between those, relative to the smaller side
	double getRelativeShift(const BBox& other) const;
	// get overlap of two areas
	double overlap(const BBox& other) const;
	
	void make_counter_clockwise();
	bool is_counter_clockwise() const;
};

// a pose of the target, consisting of its centre, log size, and orientation (rotation)
struct Pose4 {
	double data[4];
	
	Pose4(void) {} // empty just for vectors etc.
	Pose4(const BBox& ta) {
		data[0] = ta[0].x + 0.5 * (ta[0].x + ta[2].x);
		data[1] = ta[0].y + 0.5 * (ta[0].y + ta[2].y);
		data[2] = log(ta.getShorterSide()) - .5; // no idea why -.5, but it works, so don't touch!
		cv::Point2d vec0to3 = ta.p3 - ta.p0;
		data[3] = atan2(vec0to3.y, vec0to3.x); // in the -π:π range
	}
	
	double& operator[](unsigned index) {
		return data[index];
	}
	const double& operator[](unsigned index) const {
		return data[index];
	}
	
	friend std::ostream& operator<<(std::ostream& stream, const Pose4& pose);

};

// get an interpolated pixel value
cv::Vec3b interpolateColour(const cv::Mat& image, const cv::Point2d& point);
double interpolateGrayscale(const cv::Mat& image, const cv::Point2d& point);
// extract some information from images
void gradientFromBGR8u(const cv::Mat& image, Gradient& gr);
void evidenceFromBGR8u(const cv::Mat& image, const Gradient& gr, Evidence& ev);
// probabilistic stuff
double getPoseProbabilityCircRot(Pose4 pose, const Pose4& poseMean, const double * poseCov);
// comparator function for sorting 2D points
bool comparePoints(const cv::Point& p1, const cv::Point& p2);



}


#endif // __CVUTILS_H__

