#include "ransac.h"

using namespace cv;

namespace ltflo {


/* Lines RANSAC for SIMILARITY estimation
All the matrices are stored column-wise!
output arguments:
A              -o- 3*3    - similarity matrix
inliers        -o- 1*LEN  - 0/1
D              -o- 1*LEN  - errors of correspondences
score          -o- scalar - evidence of returned transf.
input arguments:
u              -i- 13*LEN - correspondences in homogeneous coordinates,
                            lines, points and index together, in form [l1;l2;x1;x2;ind]
evidDist       -i- M*N    - inverted Chamfer map
evidAng        -i- M*N    - angle of nearest edge
win            -i- 2*4    - tracked area
thr            -i- scalar - RANSAC's ilier/outlier error threshold
conf           -i- scalar - RANSAC's terminating confidence
expScaleChange -i- scalar - expected change of scale
verbose        -i- scalar - verbose mode
M              -i- scalar - image height
N              -i- scalar - image width
LEN            -i- scalar - number of correspondences

*/
void ransac(double * A, bool * inliers, double * D, double * score,
			double * u, const double * evidDist, const double * evidAng,
			const double * winOld, const double * win, double thr, double conf, double expScaleChange, bool verbose,
			unsigned M, unsigned N, unsigned LEN, bool repair) {

	unsigned i;
	unsigned maxSam = MAX_SAM, sam, I = 0, sample[3], mss = 3;
	double sc, scS = 0;
	bool newMax = false;
	double a[3*3], *d, *buffer;
	
//	const Mat evidDistM(N, M, CV_64F, (double*)evidAng);
//	Mat evidDistT = evidDistM.t();
//	normalize(evidDistT, evidDistT, 0, 1, NORM_MINMAX);
//	namedWindow("EvidD", WINDOW_NORMAL);
//	imshow("EvidD", evidDistT(Range(400,800), Range(800,1200)));
//	waitKey(0);

	A[0] = A[4] = A[8] = 1;
	A[1] = A[2] = A[3] = A[5] = A[6] = A[7] = 0;
	for (i = 0; i < LEN; ++i) {
		inliers[i] = 0;
		D[i] = 1e16;
	}
	*score = 0;
	
	if (LEN < 2) {
		return;
	}
	
	d = (double *) malloc(LEN * sizeof(double));
	buffer = (double *) malloc(24 * LEN * sizeof(double));
	
	thr = thr * thr;
	
	qsort(u, LEN, U_SIZE * sizeof(double), &compareAngle);
	
	sam = 0;
	
	while (sam < maxSam) {
		++sam;
		if (verbose && !(sam % 100)) {
			printf("Sample #%u...\n", sam);
		}
		
		/*---=== SAMPLE ===---*/
		getRandTriangle(sample, LEN, 1);
		
		/*---=== MODEL ===---*/
		triangle2A(a, u, sample, 1, U_SIZE);
		if (det3(a) < DET_TOL) {
			if (verbose) {
				printf("A close to singular, not using it.\n");
			}
			continue;
		}
		/*printf("a:");
		for (i = 0; i < 9; ++i) {
			if (!(i % 3)) {
				printf("\n");
			}
			printf("\t%7.2f", a[i%3*3 + i/3]);
		}
		printf("\n");*/
		
		/*---=== CONSENSUS/SCORE ===---*/
		sc = evidenceScore(a, u, evidDist, evidAng, expScaleChange, winOld, win, M, N, LEN, U_SIZE, repair);
		/*printf("score = %.3f\n", sc);*/
		if (*score < sc) {
			aLineDistSqU(D, a, u, LEN, win, U_SIZE, buffer);
//			aErrorSqU(D, a, u, NULL, LEN, U_SIZE, buffer);
			I = getInliers(inliers, D, thr, LEN);
			if(I>=2)
			{
				*score = sc;
				memcpy(A, a, 3*3 * sizeof(double));
				newMax = true;
			}
			else if (verbose) {
				printf("Rejected a new high score solution with only %i inliers.\n",I);
			}
		}
		
		/*---=== LOCAL OPTIMISATION ===---*/
		if (scS < sc && sc > *score * .5) { /* PLUS: at least half of optimal score */
			if (verbose) {
				printf("%.2f: Going into LO!\n", sc);
			}
			scS = sc;
			if (newMax) { /*we already have d, just copy it*/
				memcpy(d, D, LEN * sizeof(double));
			} else { /*we need to compute d*/
				aLineDistSqU(d, a, u, LEN, win, U_SIZE, buffer);
//				aErrorSqU(d, a, u, NULL, LEN, U_SIZE, buffer);
			}
			
			sc = localOptimisation(a, d, u, thr, evidDist, evidAng, winOld, win, expScaleChange, M, N, LEN, buffer, repair, verbose);
			
			if (*score < sc) {
				aLineDistSqU(D, a, u, LEN, win, U_SIZE, buffer); /*SPEED-UP: maybe get D from LO procedure?*/
//				aErrorSqU(D, a, u, NULL, LEN, U_SIZE, buffer);
				I = getInliers(inliers, D, thr, LEN);
				if(I>=2)
				{
					*score = sc;
					memcpy(A, a, 3*3 * sizeof(double));
					newMax = true;
					if (verbose) {
						printf("LO helped - from %.2f to %.2f.\n", scS, sc);
					}
				}
				else if (verbose) {
					printf("Rejected the LO solution despite high score because it has only %i inliers.\n",I);
				}
			}
		}
		
		/*---=== TERMINATION ===---*/
		if (newMax) {
			maxSam = std::min(MAX_SAM, nSamples(LEN, std::max(mss, I), mss, conf));
			if (verbose) {
				printf("New maximum (%.2f) from sample #%u, %u inliers, end is coming at #%u.\n", *score, sam, I, maxSam);
			}
			newMax = false;
		}
		
	}
	
	if (verbose) {
		printf("Line-evid RANSAC found evidence %.2f in %u samples. Scale change is %.2f.\n", *score, sam, getScaleChange(A));
	}
	
	free(d);
	free(buffer);
}


/*Perform local optimisation according to Lebeda, Matas, Chum: Fixing the Locally Optimized RANSAC, BMVC 2012*/
double localOptimisation(double * A,
			double * d, const double * u, double thr, const double * evidDist, const double * evidAng,
			const double * winOld, const double * win, double expScaleChange,
			unsigned M, unsigned N, unsigned len, double *buffer, bool repair, bool verbose) {
	
	double score, maxScore;
	unsigned I, Itmp, ss, iter, i;
	unsigned * inliers, * inliers_tmp, * sample;
	double a[3*3];
	
	inliers = (unsigned *) malloc (len * sizeof(unsigned));
	inliers_tmp = (unsigned *) malloc (len * sizeof(unsigned));
	
	A[0] = A[4] = A[8] = 1;
	A[1] = A[2] = A[3] = A[5] = A[6] = A[7] = 0;
	maxScore = 0;
	
	I = getInlierList(inliers, d, 2*thr, len);
	
	if (I <= 3) {
		free(inliers);
		return 0;
	} else {
		ss = std::min((unsigned)ceil(I / 10.0), 5u);
	}
	
	sample = (unsigned *) malloc (3 * ss * sizeof(unsigned));
	
	for (iter = 0; iter < INNER_RSC; ++iter) {
		getRandTriangle(sample, I, ss);
		for (i = 0; i < 3 * ss; ++i) {
			sample[i] = inliers[sample[i]];
		}
		triangle2A(a, u, sample, ss, U_SIZE);
		score = evidenceScore(a, u, evidDist, evidAng, expScaleChange, winOld, win, M, N, len, U_SIZE, repair);
		if (maxScore < score) {
			aLineDistSqU(d, a, u, len, win, U_SIZE, buffer);
			Itmp = getInlierList(inliers_tmp, d, 2*thr, len);
			if(Itmp>3)
			{
				maxScore = score;
				memcpy(A, a, 3*3 * sizeof(double));
				I=Itmp;
				memcpy(inliers, inliers_tmp, len * sizeof(unsigned));
			}
			else if (verbose)
				printf("Rejected high score solution inside LO as it only had %i inliers\n",Itmp);
		}
//		printf("Inner sample #%u, direct score %.2f, ", iter+1, score);
		
//		aErrorSqU(d, a, u, NULL, len, U_SIZE, buffer);
		score = iterativeLSq(a, d, u, thr/2, thr, 2, evidDist, evidAng, winOld, win, expScaleChange, M, N, len, buffer, repair);
		aLineDistSqU(d, a, u, len, win, U_SIZE, buffer);
		Itmp = getInlierList(inliers_tmp, d, 2*thr, len);
		if (maxScore < score) {
			if(Itmp >3) {
			maxScore = score;
			memcpy(A, a, 3*3 * sizeof(double));
				I=Itmp;
				memcpy(inliers, inliers_tmp, len * sizeof(unsigned));
			}
			else if (verbose)
				printf("Rejected high score solution from iterativeLsq inside LO as it only had %i inliers\n",Itmp);
		}
//		printf("refined score %.2f.\n", score);
	}
	
	free(sample);
	free(inliers);
	return maxScore;
}


/*performs iterative least squares with narrowing threshold from thrs to thr*/
double iterativeLSq(double * A,
			double * d, const double * u, double thr, double thrs, unsigned iters,
			const double * evidDist, const double * evidAng, const double * winOld, const double * win,
			double expScaleChange, unsigned M, unsigned N, unsigned len, double *buffer, bool repair) {
	
	double score, maxScore = 0;
	double a[3*3];
	double dth = (thrs - thr) / (iters - 1);
	unsigned * inliers;
	unsigned I, i;
	
	inliers = new unsigned[len];
	
/*	A[0] = A[4] = A[8] = 1;*/
/*	A[1] = A[2] = A[3] = A[5] = A[6] = A[7] = 0;*/
	
/*	I = getInliers(inliers, d, thr, len);*/
/*	if (I < 2) {*/
/*		free(inliers);*/
/*		return 0;*/
/*	}*/
	
/*	l2aNLnm(A, u, I, inliers, win, len, U_SIZE, buffer);*/
	memcpy(a, A, 3*3 * sizeof(double));
	maxScore = evidenceScore(A, u, evidDist, evidAng, expScaleChange, winOld, win, M, N, len, U_SIZE, repair);
	
	for (i = 0; i < iters; ++i) {
		I = getInlierList(inliers, d, thrs, len);
		if (I < 2) {
			delete[] inliers;
			return maxScore;
		}
//		l2aNLnm(a, u, I, inliers, win, len, U_SIZE, buffer);
//		l2aSimon(a, u, I, inliers, win, U_SIZE, buffer);
		// this one is the best:
//		l2aSimonOrig(a, u, I, inliers, win, U_SIZE, buffer);
		// ^^ that one ^^
//		l2aEvid(a, u, I, inliers, evidDist, evidAng, M, N, U_SIZE, repair);
		if (det3(a) < DET_TOL) {
			continue;
		}
		aLineDistSqU(d, a, u, len, win, U_SIZE, buffer);
//		aErrorSqU(d, a, u, NULL, len, U_SIZE, buffer);
		
		score = evidenceScore(a, u, evidDist, evidAng, expScaleChange, winOld, win, M, N, len, U_SIZE, repair);
		
		if (maxScore < score) {
			maxScore = score;
			memcpy(A, a, 3*3 * sizeof(double));
		}
		thrs = thrs - dth;
	}
	
	
	delete[] inliers;
	return maxScore;
}






// outer thing, includes the ransac and refinement, returns number of inliers
unsigned ransacWrapper(cv::Mat& A, std::vector<double>& d, double& score, double& optScore,
					std::vector<double>& linesOld, std::vector<double>& linesNew, std::vector<unsigned>& correspFull,
					const Evidence& evidence, const cv::Mat& weightsField, const BBox& winOld, const BBox& winNew,
					double errThr, double conf, double scaleChange, bool repair, bool verbose, const cv::Mat& image) {
	
	assert(correspFull.size() % 4 == 0);
	assert(linesOld.size() % 3 == 0);
	assert(linesNew.size() % 3 == 0);
	
	unsigned LEN = correspFull.size()/4;
	
	Mat u(0, LEN, CV_64F);
	// and now let's fill it!
	Mat l1(LEN, 3, CV_64F, &linesOld[0]); // create lightweight matrix with the lines (transposed, row-major order!)
	u.push_back(l1.t());
	// the same for others
	Mat l2(LEN, 3, CV_64F, &linesNew[0]);
	u.push_back(l2.t());
	Mat uX(LEN, 4, CV_32S, &correspFull[0]); // signed, but no problem expected
	uX.convertTo(uX, CV_64F);
	uX = uX.t() + 1; // to convert to 1-based indexing!
	u.push_back(uX.row(0)); // x1
	u.push_back(uX.row(1)); // y1
	u.push_back(Mat::ones(1, LEN, CV_64F)); // 1's
	u.push_back(uX.row(2)); // x2
	u.push_back(uX.row(3)); // y2
	u.push_back(Mat::ones(2, LEN, CV_64F)); // 1's
	for (unsigned i = 0; i < LEN; ++i) { u.at<double>(12,i) = i; }
	
	u = u.t(); // to get the data in column-major order, as expected in RANSAC, which wants each "correspondence" stored in memory together, locally
	
	// output data
	double* AD = (double*) A.ptr();
	bool inliers[LEN];
	// input data
	double* uD = (double*) u.ptr();
	Mat evidDistT = evidence.d.t();
	double* evidDistD = (double*) evidDistT.ptr();
	Mat evidAngT = evidence.a.t();
	double* evidAngD = (double*) evidAngT.ptr();
	double winOldD[2*4], winNewD[2*4];
	winOld.toDouble(winOldD);
	winNew.toDouble(winNewD);
	
	
	if (!A.isContinuous() || !u.isContinuous() || !evidDistT.isContinuous() || !evidAngT.isContinuous()) {
		throw "Can't get the data from a non-continuous matrix!";
	}
	
	if (verbose) {
		std::cout << "No LO for now!\n";
	}
	ransac(AD, inliers, &d[0], &score,
			uD, evidDistD, evidAngD,
			winOldD, winNewD, errThr, conf, scaleChange, verbose,
			evidence.a.rows, evidence.a.cols, LEN, repair);
	
	// refinement: weights field maximisation
	Mat Acopy = A.t(); // here we go back to row-major
	Mat u2(uX, Range(2, 4), Range::all()); // 3rd and 4th row contain u2
	u2 -= 1.0; // conver to 0-based again
	
	refineA(Acopy, &optScore, u2, weightsField); // this one uses 0-based indexing
	if (!Acopy.isContinuous()) {
		throw "Can't get the data from a non-continuous matrix!";
	}
	Acopy = Acopy.t(); // and back to column-major!
	double refinedScore = evidenceScore((double *)Acopy.ptr(), uD, evidDistD, evidAngD, // uD used 1-based indexing
							scaleChange, winOldD, winNewD, evidence.a.rows, evidence.a.cols, LEN, 13, repair);
	
	if (verbose) {
		std::cout << "A_RANSAC:\n" << A.t() << "\n";
		std::cout << "A_REFINED:\n" << Acopy.t() << "\n";
	}
	
	if (refinedScore >= score * 0.7) {
		if (verbose) {
			std::cout << "Refinement accepted! " << refinedScore << ">" << score << "*.7!\n";
		}
		score = refinedScore;
		A = Acopy;
	} else {
		if (verbose) {
			std::cout << "Refinement rejected! " << refinedScore << "<" << score << "*.7!\n";
		}
	}
	
	aLineDistSq(&d[0], (double*) A.ptr(), &linesOld[0], &linesNew[0], LEN, winNewD, NULL);
	
	A = A.t(); // back from col-major to row-major order
	
	unsigned numInliers = 0;
	for (unsigned i = 0; i < LEN; ++i) {
		d[i] = sqrt(d[i]);
		if (d[i] <= errThr) {
			++numInliers;
		}
	}
	
	optScore *= 10;
	
	return numInliers;
}













} // namespace ltflo


