#include "ransacTools.h"

using std::min;
using std::max;

namespace ltflo {

/*Compare angles of l1 modulo PI, both parameters are joint correspondences [l1;l2;l3;l4]*/
int compareAngle(const void * u1, const void * u2) {
	double * c1 = (double *) u1;
	double * c2 = (double *) u2;
	double a1 = doubleMod(atan2(c1[2], c1[1]) , M_PI);
	double a2 = doubleMod(atan2(c2[2], c2[1]) , M_PI);
	if (a1 > a2) {
		return 1;
	}
	if (a1 < a2) {
		return -1;
	}
	return 0;
}

/*Number of samples to report success of 'ss'-point RANSAC with 'inl' inliers out of 'len' points with confidence 'conf'*/
unsigned nSamples(unsigned len, unsigned inl, unsigned ss, double conf) {
	unsigned i;
	double p = 1;
	double EPS = 1e-15;
	for (i = 0; i < ss; ++i) {
		p *= ((double)inl - i) / (len - i);
	}
	if (p < EPS) {
		return 4e9;
	}
	if (p > 1 - EPS) {
		return 0;
	}
	return (unsigned) ceil(log(1 - conf) / log(1 - p));
}

/*Generate samples = triplets of lines with large mutual angles*/
void getRandTriangle(unsigned * sample, unsigned len, unsigned noTriangles) {
	double sigma = len / 12.0;
	unsigned i, ind;
//		std::cerr<<"triangle len="<<len<<std::endl;
	for (i = 0; i < noTriangles; ++i) {
		ind = rand() % len;
		sample[i*3] = ind;
		ind += (unsigned)round(min(len*.45, max(len*.2, len / 3.0 + randNorm() * sigma)));
		ind %= len;
		sample[i*3+1] = ind;
		ind += (unsigned)round(min(len*.45, max(len*.2, len / 3.0 + randNorm() * sigma)));
		ind %= len;
		sample[i*3+2] = ind;
	}
}

/*return binary field with 1's at positions of points with error lower than threshold*/
unsigned getInliers(bool * inliers, double * d, double thr, unsigned len) {
	unsigned i, I = 0;
	for (i = 0; i < len; ++i) {
		if (d[i] <= thr) {
			inliers[i] = 1;
			++I;
		} else {
			inliers[i] = 0;
		}
	}
	return I;
}

/*finds positions of points with error lower than threshold, inlier count returned*/
unsigned getInlierList(unsigned * inliers, double * d, double thr, unsigned len) {
	unsigned i, I = 0;
	for (i = 0; i < len; ++i) {
		if (*d++ <= thr) {
			*inliers++ = i;
			++I;
		}
	}
	return I;
}


/*computes evidence of a transformation*/
double evidenceScore(double * A, const double * u, const double * evidDist, const double * evidAng,
					double expScaleChange, const double * winOld, const double * win,
					unsigned M, unsigned N, unsigned len, unsigned u_len, bool repair) {
	double score = 0, wLoc, wAng;
	double scaleChange, regularisation, regularisation2;
	double rot = getRotation(A);
	double x, y, w;
	unsigned i, bi;
	point pa[4], pb[4];
	
	scaleChange = getScaleChange(A);
	regularisation = min(scaleChange / expScaleChange, expScaleChange / scaleChange);
	
	for (i = 0; i < 4; ++i) {
		pa[i].x = *winOld++;
		pa[i].y = *winOld++;
		x = A[0]*win[2*i] + A[3]*win[2*i+1] + A[6];
		y = A[1]*win[2*i] + A[4]*win[2*i+1] + A[7];
		w = A[2]*win[2*i] + A[5]*win[2*i+1] + A[8];
		pb[i].x = x / w;
		pb[i].y = y / w;
	}
	regularisation2 = intersectionArea(4, pa, 4, pb) / (rectangleArea(pa) + rectangleArea(pb) - intersectionArea(4, pa, 4, pb));
	regularisation2 = min(1., .25 + 1.5 * regularisation2);
	
	
	for (i = 0; i < len; ++i) {
		x = A[0]*u[6] + A[3]*u[7] + A[6]*u[8];
		y = A[1]*u[6] + A[4]*u[7] + A[7]*u[8];
		w = A[2]*u[6] + A[5]*u[7] + A[8]*u[8];
		x = max(1., min(double(N), x / w)) - 1; /*coordinates to indices!*/
		y = max(1., min(double(M), y / w)) - 1;
		bi = (unsigned)(y+.5) + M * (unsigned)(x+.5);
		if (repair) {
			wAng = fabs(cos(atan2(u[1], u[0]) + rot - evidAng[bi]));
		} else {
			wAng = .5 + .5 * cos(atan2(u[1], u[0]) + rot - evidAng[bi]);
		}
		wLoc = evidDist[bi];
		score += wAng * wLoc;
		u += u_len;
	}
	
	score /= len;
	score *= (regularisation) * regularisation2;
	
	return score;
}

/*computes a similarity transformation from T triplets of line intersections (triangles vetrices), return error code
error codes:
 0 - success
 1 - degenerate sample*/
int triangle2A(double * A, const double * u, const unsigned * indices, unsigned noTriangles, unsigned u_len) {
	double * eqL, * eqR;
	unsigned i, j;
	double p1[3], p2[3], x1, y1, x2, y2;
	double v[4];
	int res;
	
	
	eqL = (double *) malloc(4 * 2 * 3 * noTriangles * sizeof(double)); /*4 columns, 2 eqs per corr, 3 corrs per triangle*/
	eqR = (double *) malloc(2 * 3 * noTriangles * sizeof(double));/*the same with one column*/
	
	for (i = 0; i < noTriangles; ++i) {
		for (j = 0; j < 3; ++j) {
			crossU(p1, u, indices[j], indices[(j + 1) % 3], 0, u_len); /*get first point corresp*/
			crossU(p2, u, indices[j], indices[(j + 1) % 3], 1, u_len);
			x1 = p1[0] / p1[2];
			y1 = p1[1] / p1[2];
			x2 = p2[0] / p2[2];
			y2 = p2[1] / p2[2];
			/*printf("p1=[%.1f; %.1f], p2=[%.1f; %.1f]\n", x1, y1, x2, y2);*/
		
			/*//TODO norm!?! /100 not enough?*/
		
			/*eqL = [x1(1,:)', -x1(2,:)', ones(len,1), zeros(len,1);
				   x1(2,:)',  x1(1,:)', zeros(len,1), ones(len,1)];*/
			eqL[2*3*noTriangles*0 + 2*3*i + 2*j] = x1;
			eqL[2*3*noTriangles*1 + 2*3*i + 2*j] = -y1;
			eqL[2*3*noTriangles*2 + 2*3*i + 2*j] = 100;
			eqL[2*3*noTriangles*3 + 2*3*i + 2*j] = 0;
			eqL[2*3*noTriangles*0 + 2*3*i + 2*j+1] = y1;
			eqL[2*3*noTriangles*1 + 2*3*i + 2*j+1] = x1;
			eqL[2*3*noTriangles*2 + 2*3*i + 2*j+1] = 0;
			eqL[2*3*noTriangles*3 + 2*3*i + 2*j+1] = 100;
			
			eqR[2*3*i + 2*j] = x2;
			eqR[2*3*i + 2*j+1] = y2;
		}
		/*printf("eqL\n=====================\n");
		for (j = 0; j < 2*3*4; ++j) {
			printf(" %7.1f", eqL[j%4*(2*3*noTriangles) + j/4]);
			if (j % 4 == 3) {
				printf("\n");
			}
		}
		printf("eqR\n=====================\n");
		for (j = 0; j < 2*3; ++j) {
			printf(" %7.1f ", eqR[j]);
		}
		printf("\n");*/
		
		indices += 3;
	}
	
	/*least squares solution to eqL . v = eqR*/
	res = lsqPInv(v, eqL, eqR, 2*3*noTriangles, 4);
	
	/*printf("v: [%.2f; %.2f; %.3f; %.3f]\n", v[0], v[1], v[2], v[3]);*/
	
	A[0] = A[4] = v[0];
	A[1] = v[1];
	A[2] = A[5] = 0;
	A[3] = -v[1];
	A[6] = v[2] * 100; /*primitive normalisation*/
	A[7] = v[3] * 100;
	A[8] = 1;
	
	/*printf("A:");
	for (i = 0; i < 9; ++i) {
		if (!(i % 3)) {
			printf("\n");
		}
		printf("\t%7.2f", A[i%3*3 + i/3]);
	}
	printf("\n");*/
	
	free(eqR);
	free(eqL);
	return res;
}





} // namespace ltflo




